==> InfiniTec.Exchange_0.96_Source.zip: Security/ExchangeSecurity.cs <==
using System;
using System.Collections.Generic;
using System.Text;
using System.Security.AccessControl;
using System.Security.Principal;
using System.Runtime.InteropServices;
using InfiniTec.Security;

namespace InfiniTec.Exchange.Security
{
    /// <summary>
    /// This class provides the implementation for accessing and manipulation
    /// of Exchange security descriptors
    /// </summary>
    public abstract class ExchangeSecurity
    {
        private static Dictionary<SecurityIdentifier, AccountType> _AccountTypeCache = new Dictionary<SecurityIdentifier, AccountType>();

        private RawSecurityDescriptor _SecurityDescriptor;
        private Dictionary<SecurityIdentifier, ItemAce> _AccessControlEntries;

        /// <summary>
        /// Returns the access control entries which make up this discretionary
        /// access control list
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <returns></returns>
        internal IList<T> GetAccessControlEntries<T>() where T: ItemAce
        {
            List<T> result;

            result = new List<T>();
            foreach (T entry in _AccessControlEntries.Values)
            {
                result.Add(entry);
            }

            return result.AsReadOnly();
        }
        /// <summary>
        /// Creates a new instance of this class using the specified values
        /// </summary>
        /// <param name="buffer">A buffer containing the security descriptor</param>
        /// <param name="offset">The offset in the buffer where the security descriptor is stored</param>
        protected ExchangeSecurity(byte[] buffer, int offset)
        {
            ItemAce ace;

            _SecurityDescriptor = new RawSecurityDescriptor(buffer, offset);

            _AccessControlEntries = new Dictionary<SecurityIdentifier, ItemAce>();
            foreach (CommonAce entry in _SecurityDescriptor.DiscretionaryAcl)
            {
                if (!_AccessControlEntries.ContainsKey(entry.SecurityIdentifier))
                {
                    ace = AddAce(entry.SecurityIdentifier);
                }
                else
                {
                    ace = _AccessControlEntries[entry.SecurityIdentifier];
                }
                ace.LoadAce(entry);
            }
        }

        /// <summary>
        /// Returns the ACE with the specified identity.
        /// </summary>
        /// <param name="identity">The identity of the desired ACE</param>
        /// <param name="ace">A reference to an ItemAce</param>
        /// <returns>true, if the ACE was found. False, otherwise</returns>
        protected bool TryGetAce(SecurityIdentifier identity, out ItemAce ace)
        {
            return _AccessControlEntries.TryGetValue(identity, out ace);
        }
        /// <summary>
        /// Creates a new ACE and adds it to the access control list
        /// </summary>
        /// <param name="identity">The identity of the ace</param>
        /// <returns>The newly created access control entry</returns>
        protected ItemAce AddAce(SecurityIdentifier identity)
        {
            ItemAce ace;

            ace = CreateAce(identity);

            ace.AceCreated += new EventHandler<AceCreatedEventArgs>(Ace_AceCreated);
            _AccessControlEntries.Add((SecurityIdentifier) ace.IdentityReference, ace);

            return ace;
        }

        private void Ace_AceCreated(object sender, AceCreatedEventArgs e)
        {
            _SecurityDescriptor.DiscretionaryAcl.InsertAce(GetInsertPosition(e.Ace),
                e.Ace);
        }

        private int GetInsertPosition(CommonAce ace)
        {
            int i;

            i = 0;
            while (i < _SecurityDescriptor.DiscretionaryAcl.Count &&
                CompareAce((CommonAce) _SecurityDescriptor.DiscretionaryAcl[i], ace)  < 0) i++;

            return i;
        }

        private int CompareAce(CommonAce leftAce, CommonAce rightAce)
        {
            int result;

            if (leftAce == null && rightAce == null) return 0;
            if (leftAce == null) return -1;
            if (rightAce == null) return 1;

            result = CompareAccountType(leftAce.SecurityIdentifier, rightAce.SecurityIdentifier);

            if (result != 0) return result;

            if (IsGroup(leftAce.SecurityIdentifier))
            {
                result = CompareAceType(leftAce.AceType, rightAce.AceType);
                if (result != 0) return result;

                return (leftAce.SecurityIdentifier.CompareTo(rightAce.SecurityIdentifier));
            }
            else
            {
                if (leftAce.SecurityIdentifier == rightAce.SecurityIdentifier)
                {
                    if ((leftAce.AceFlags & AceFlags.InheritOnly) != AceFlags.InheritOnly &&
                        (rightAce.AceFlags & AceFlags.InheritOnly) == AceFlags.InheritOnly)
                        return -1;
                    else if ((leftAce.AceFlags & AceFlags.InheritOnly) == AceFlags.InheritOnly &&
                        (rightAce.AceFlags & AceFlags.InheritOnly) != AceFlags.InheritOnly)
                        return 1;
                    else
                    {
                        if ((leftAce.InheritanceFlags & InheritanceFlags.ContainerInherit) == InheritanceFlags.ContainerInherit &&
                            (rightAce.InheritanceFlags & InheritanceFlags.ContainerInherit) != InheritanceFlags.ContainerInherit)
                            return -1;
                        else if ((leftAce.InheritanceFlags & InheritanceFlags.ContainerInherit) == InheritanceFlags.ContainerInherit &&
                            (rightAce.InheritanceFlags & InheritanceFlags.ContainerInherit) != InheritanceFlags.ContainerInherit)
                            return 1;

                        return CompareAceType(leftAce.AceType, rightAce.AceType);
                    }
                }
            }

            if (leftAce.SecurityIdentifier.IsWellKnown(WellKnownSidType.WorldSid)) return 1;
            if (rightAce.SecurityIdentifier.IsWellKnown(WellKnownSidType.WorldSid)) return -1;

            return 0;
        }

        private int CompareAccountType(SecurityIdentifier leftSid, SecurityIdentifier rightSid)
        {
            bool leftIsGroup;
            bool rightIsGroup;

            if (leftSid.IsWellKnown(WellKnownSidType.WorldSid)) return 1;
            else if (rightSid.IsWellKnown(WellKnownSidType.WorldSid)) return -1;

            leftIsGroup = IsGroup(leftSid);
            rightIsGroup = IsGroup(rightSid);

            if (!leftIsGroup && rightIsGroup) return -1;
            else if (leftIsGroup && !rightIsGroup) return 1;
            
            return 0;
        }

        private bool IsGroup(SecurityIdentifier sid)
        {
            IdentityResolver resolver;
            AccountType type;

            if (sid.IsWellKnown(WellKnownSidType.AnonymousSid)) return false;
            if (sid.IsWellKnown(WellKnownSidType.WorldSid)) return false;

            if (_AccountTypeCache.ContainsKey(sid))
            {
                type = _AccountTypeCache[sid];
            }
            else
            {
                resolver = new IdentityResolver(sid, "w2k3server");
                type = resolver.AccountType;
                _AccountTypeCache[sid] = type;
            }

            return (type == AccountType.WellknownGroup || type == AccountType.Group);
        }

        private int CompareAceType(AceType left, AceType right)
        {
            if (left == right) return 0;

            if (left == AceType.AccessAllowed) return -1;

            return 1;
        }

        /// <summary>
        /// Returns the binary length of the security descriptor.
        /// </summary>
        public int BinaryLength
        {
            get { return _SecurityDescriptor.BinaryLength; }
        }

        /// <summary>
        /// Returns the security descriptor in SDDL form.
        /// </summary>
        /// <param name="sections">The <see cref="AccessControlSections"/> to return</param>
        /// <returns>The requested sections in SDDL form</returns>
        public string GetSddlForm(AccessControlSections sections)
        {
            return _SecurityDescriptor.GetSddlForm(sections);
        }

        /// <summary>
        /// Set the <see cref="ControlFlags"/> of the security descriptor
        /// </summary>
        /// <param name="flags">The new <see cref="ControlFlags"/></param>
        public void SetFlags(ControlFlags flags)
        {
            _SecurityDescriptor.SetFlags(flags);
        }

        /// <summary>
        /// Writes the security descriptor to the buffer
        /// </summary>
        /// <param name="buffer">The buffer for the security descriptor</param>
        /// <param name="offset">The offset in the buffer where to begin the write operation</param>
        public void GetBinaryForm(byte[] buffer, int offset)
        {
            _SecurityDescriptor.GetBinaryForm(buffer, offset);
        }

        /// <summary>
        /// Returns the <see cref="ControlFlags"/> of the security descriptor.
        /// </summary>
        public ControlFlags ControlFlags
        {
            get { return _SecurityDescriptor.ControlFlags; }
        }

        /// <summary>
        /// Gets or sets the primary group of the item
        /// </summary>
        public SecurityIdentifier Group
        {
            get { return _SecurityDescriptor.Group; }
            set { _SecurityDescriptor.Group = value; }
        }

        /// <summary>
        /// Gets or sets the owner of the item
        /// </summary>
        public SecurityIdentifier Owner
        {
            get { return _SecurityDescriptor.Owner; }
            set { _SecurityDescriptor.Owner = value; }
        }

        /// <summary>
        /// Sets the specified access for the identity
        /// </summary>
        /// <param name="identity">The identity for wich to set the permissions</param>
        /// <param name="accessMask">The desired level of access</param>
        public virtual void SetAccess(SecurityIdentifier identity, AccessMasks accessMask)
        {
            ItemAce ace;

            if (!TryGetAce(identity, out ace))
            {
               ace = AddAce(identity);
            }

            ace.SetAccess(accessMask);
        }

        /// <summary>
        /// Removes the access for the specified identity
        /// </summary>
        /// <param name="identity">A <see cref="SecurityIdentifier"/> instance.</param>
        public void RemoveAccess(SecurityIdentifier identity)
        {
            ItemAce ace;
            List<CommonAce> aces;

            ace = _AccessControlEntries[identity];
            
            _AccessControlEntries.Remove(identity);

            aces = new List<CommonAce>(ace.GetAces());

            for (int i = _SecurityDescriptor.DiscretionaryAcl.Count - 1; i >= 0 ; i--)
			{
                if (aces.Contains((CommonAce) _SecurityDescriptor.DiscretionaryAcl[i]))
                    _SecurityDescriptor.DiscretionaryAcl.RemoveAce(i);
			}
        }

        /// <summary>
        /// Creates a new access control entry for the specified identity.
        /// </summary>
        /// <param name="identity">A <see cref="SecurityIdentifier"/> instance</param>
        /// <returns>The newly create <see cref="ItemAce"/> instance.</returns>
        protected virtual ItemAce CreateAce(SecurityIdentifier identity)
        {
            return new ItemAce(identity);
        }
    }
}

==> InfiniTec.Exchange_0.97_Source.zip: Security/ExchangeSecurity.cs <==
using System;
using System.Collections.Generic;
using System.Security.AccessControl;
using System.Security.Principal;

namespace InfiniTec.Exchange.Security
{
    /// <summary>
    /// This class provides the implementation for accessing and manipulation
    /// of Exchange security descriptors
    /// </summary>
    public abstract class ExchangeSecurity
    {
        private RawSecurityDescriptor _SecurityDescriptor;
        private Dictionary<SecurityIdentifier, ItemAce> _AccessControlEntries;
        private Connection _Connection;

        /// <summary>
        /// Returns the access control entries which make up this discretionary
        /// access control list
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <returns></returns>
        internal IList<T> GetAccessControlEntries<T>() where T: ItemAce
        {
            List<T> result;

            result = new List<T>();
            foreach (T entry in _AccessControlEntries.Values)
            {
                result.Add(entry);
            }

            return result.AsReadOnly();
        }
        /// <summary>
        /// Creates a new instance of this class using the specified values
        /// </summary>
        /// <param name="buffer">A buffer containing the security descriptor</param>
        /// <param name="offset">The offset in the buffer where the security descriptor is stored</param>
        /// <param name="connection">The connection used with this security descriptor</param>
        protected ExchangeSecurity(byte[] buffer, int offset, Connection connection)
        {
            ItemAce ace;

            _Connection = connection;
            _SecurityDescriptor = new RawSecurityDescriptor(buffer, offset);

            _AccessControlEntries = new Dictionary<SecurityIdentifier, ItemAce>();
            foreach (CommonAce entry in _SecurityDescriptor.DiscretionaryAcl)
            {
                if (!_AccessControlEntries.ContainsKey(entry.SecurityIdentifier))
                {
                    ace = AddAce(entry.SecurityIdentifier);
                }
                else
                {
                    ace = _AccessControlEntries[entry.SecurityIdentifier];
                }
                ace.LoadAce(entry);
            }
        }

        /// <summary>
        /// Returns the ACE with the specified identity.
        /// </summary>
        /// <param name="identity">The identity of the desired ACE</param>
        /// <param name="ace">A reference to an ItemAce</param>
        /// <returns>true, if the ACE was found. False, otherwise</returns>
        protected bool TryGetAce(SecurityIdentifier identity, out ItemAce ace)
        {
            return _AccessControlEntries.TryGetValue(identity, out ace);
        }
        /// <summary>
        /// Creates a new ACE and adds it to the access control list
        /// </summary>
        /// <param name="identity">The identity of the ace</param>
        /// <returns>The newly created access control entry</returns>
        protected ItemAce AddAce(SecurityIdentifier identity)
        {
            ItemAce ace;

            ace = CreateAce(identity);

            ace.AceCreated += new EventHandler<AceCreatedEventArgs>(Ace_AceCreated);
            _AccessControlEntries.Add(ace.IdentityReference, ace);

            return ace;
        }

        private void Ace_AceCreated(object sender, AceCreatedEventArgs e)
        {
            _SecurityDescriptor.DiscretionaryAcl.InsertAce(GetInsertPosition(e.Ace),
                e.Ace);
        }

        /// <summary>
        /// Sorts the discretionary access control list so that it is in Exchange canonical order.
        /// </summary>
        protected void SortAcl()
        {
            CommonAce[] acl;

            acl = new CommonAce[_SecurityDescriptor.DiscretionaryAcl.Count];
            
            _SecurityDescriptor.DiscretionaryAcl.CopyTo(acl, 0);
            _SecurityDescriptor.DiscretionaryAcl = new RawAcl(RawAcl.AclRevision, 0);

            Array.Sort(acl, new AceComparer(_Connection));

            for (int i = 0; i < acl.Length; i++)
            {
                _SecurityDescriptor.DiscretionaryAcl.InsertAce(i, acl[i]);                
            }
        }

        private int GetInsertPosition(CommonAce ace)
        {
            int i;
            AceComparer comparer;

            comparer = new AceComparer(_Connection);

            i = 0;

            while (i < _SecurityDescriptor.DiscretionaryAcl.Count &&
                comparer.Compare((CommonAce) _SecurityDescriptor.DiscretionaryAcl[i], ace)  < 0) i++;

            return i;
        }

     
        /// <summary>
        /// Returns the binary length of the security descriptor.
        /// </summary>
        public int BinaryLength
        {
            get { return _SecurityDescriptor.BinaryLength; }
        }

        /// <summary>
        /// Returns the security descriptor in SDDL form.
        /// </summary>
        /// <param name="sections">The <see cref="AccessControlSections"/> to return</param>
        /// <returns>The requested sections in SDDL form</returns>
        public string GetSddlForm(AccessControlSections sections)
        {
            return _SecurityDescriptor.GetSddlForm(sections);
        }

        /// <summary>
        /// Set the <see cref="ControlFlags"/> of the security descriptor
        /// </summary>
        /// <param name="flags">The new <see cref="ControlFlags"/></param>
        public void SetFlags(ControlFlags flags)
        {
            _SecurityDescriptor.SetFlags(flags);
        }

        /// <summary>
        /// Writes the security descriptor to the buffer
        /// </summary>
        /// <param name="buffer">The buffer for the security descriptor</param>
        /// <param name="offset">The offset in the buffer where to begin the write operation</param>
        public void GetBinaryForm(byte[] buffer, int offset)
        {
            _SecurityDescriptor.GetBinaryForm(buffer, offset);
        }

        /// <summary>
        /// Returns the <see cref="ControlFlags"/> of the security descriptor.
        /// </summary>
        public ControlFlags ControlFlags
        {
            get { return _SecurityDescriptor.ControlFlags; }
        }

        /// <summary>
        /// Gets or sets the primary group of the item
        /// </summary>
        public SecurityIdentifier Group
        {
            get { return _SecurityDescriptor.Group; }
            set { _SecurityDescriptor.Group = value; }
        }

        /// <summary>
        /// Gets or sets the owner of the item
        /// </summary>
        public SecurityIdentifier Owner
        {
            get { return _SecurityDescriptor.Owner; }
            set { _SecurityDescriptor.Owner = value; }
        }

        /// <summary>
        /// Sets the specified access for the identity
        /// </summary>
        /// <param name="identity">The identity for wich to set the permissions</param>
        /// <param name="accessMask">The desired level of access</param>
        public virtual void SetAccess(SecurityIdentifier identity, AccessMasks accessMask)
        {
            ItemAce ace;

           UpdateControlFlagsForModification();

            if (!TryGetAce(identity, out ace))
            {
               ace = AddAce(identity);
            }

            ace.SetAccess(accessMask);
  
            SortAcl();
        }

        /// <summary>
        /// Updates the control flags for modification.
        /// </summary>
        protected virtual void UpdateControlFlagsForModification()
        {
            SetFlags(ControlFlags & ~(ControlFlags.DiscretionaryAclAutoInherited | ControlFlags.DiscretionaryAclDefaulted));
        }

        /// <summary>
        /// Removes the access for the specified identity
        /// </summary>
        /// <param name="identity">A <see cref="SecurityIdentifier"/> instance.</param>
        public virtual void RemoveAccess(SecurityIdentifier identity)
        {
            CommonAce ace;

            UpdateControlFlagsForModification();         

            _AccessControlEntries.Remove(identity);

            for (int i = _SecurityDescriptor.DiscretionaryAcl.Count - 1; i >= 0; i--)
            {
                ace = (CommonAce) _SecurityDescriptor.DiscretionaryAcl[i];
                if (identity != ace.SecurityIdentifier) continue;

                _SecurityDescriptor.DiscretionaryAcl.RemoveAce(i);
            }
        }

        /// <summary>
        /// Creates a new access control entry for the specified identity.
        /// </summary>
        /// <param name="identity">A <see cref="SecurityIdentifier"/> instance</param>
        /// <returns>The newly create <see cref="ItemAce"/> instance.</returns>
        protected virtual ItemAce CreateAce(SecurityIdentifier identity)
        {
            return new ItemAce(identity);
        }
    }
}
